import { reactive, isProxy, toRaw } from 'vue';

const $global = {};

function setGlobalVariable(key, value, root = '$cvt') {
  if (key) {
    $global[root][key] = value;
  } else {
    $global[root] = value;
  }
}

function getGlobalVariable(key, root = '$cvt') {
  return (key ? $global[root]?.[key] : $global[root]) || {}
}

const isUniApp = () => typeof uni !== 'undefined';

const _cloneDeep = (o) => {
  const { _ } = getGlobalVariable('libs');
  return _.cloneDeep(o);
};

const _get = (v, k) => {
  const { _ } = getGlobalVariable('libs');
  return _.get(v, k);
};

const _set = (v, k, o) => {
  const { _ } = getGlobalVariable('libs');
  return _.set(v, k, o);
};

const _find = (v, f) => {
  const { _ } = getGlobalVariable('libs');
  return _.find(v, f);
};

function isObjHasBlank(obj, skipAttr = []) {
  for (let i in obj) {
    if (!obj[i] && !skipAttr.includes(i)) {
      return true;
    }
  }
}

function getObjVal(s, k, d = "--") {
  const { _ } = getGlobalVariable('libs');
  return _.get(s, k, d);
}

function toStringify(o) {
  return JSON.stringify(o);
}


function toParse(s) {
  return JSON.parse(s);
}

function uuid() {
  var s = [];
  var hexDigits = "0123456789abcdef";
  for (var i = 0; i < 36; i++) {
    s[i] = hexDigits.substr(Math.floor(Math.random() * 0x10), 1);
  }
  s[14] = "4";
  s[19] = hexDigits.substr((s[19] & 0x3) | 0x8, 1);
  s[8] = s[13] = s[18] = s[23] = "-";
  var uuid = s.join("");
  return uuid;
}

function genArrayByLength(l) {
  return Array.from({ length: l }, (v, i) => i);
}

const stringifyQuery = (params) => {
  let str = "";
  let i = 0;
  for (let k in params) {
    let q = "";
    if (i === 0) {
      q = "?";
    } else {
      q = "&";
    }
    str = str + q + k + "=" + params[k];
    i++;
  }
  return str;
};

function mitt(n) {
  return {
    all: (n = n || new Map()),
    on: function (e, t) {
      var i = n.get(e);
      i ? i.push(t) : n.set(e, [t]);
    },
    off: function (e, t) {
      var i = n.get(e);
      i && (t ? i.splice(i.indexOf(t) >>> 0, 1) : n.set(e, []));
    },
    emit: function (e, t) {
      var i = n.get(e);
      i &&
        i.slice().map(function (n) {
          n(t);
        }),
        (i = n.get("*")) &&
        i.slice().map(function (n) {
          n(e, t);
        });
    },
  };
}

const emitter = new mitt();

const defaultConfig = {
  isUp2sever: false,
  url: ''
};
class Log4js {
  config = {}

  constructor(config) {
    this.config = Object.assign(defaultConfig, config);
  }

  debug(e) {
    const prefix = `%c
  //     __  __     ____         _       __           __    __
  //    / / / /__  / / /___     | |     / /___  _____/ /___/ /
  //   / /_/ / _ \/ / / __ \    | | /| / / __ \/ ___/ / __  / 
  //  / __  /  __/ / / /_/ /    | |/ |/ / /_/ / /  / / /_/ /  
  // /_/ /_/\___/_/_/\____/     |__/|__/\____/_/  /_/\__,_/   
  //                                             
        `;
    const suffix = `%c
  //
  // =======================================================
  //
        `;
    console.log(prefix, 'color: blue');
    console.log(e);
    console.log(suffix, 'color: red');
  }

  log(e) {
    console.log(`%c ------------- power by log4js, log start -----------`, 'color: blue');
    console.log(e);
    console.log(`%c ------------- power by log4js, log end -----------`, 'color: red');
  }

  error(e) {
    console.log(`%c ------------- power by log4js, log start -----------`, 'color: blue');
    throw e
  }
}

const getDates = (d, t = 0) => {
  var curDate = new Date(d.replace(/-/g, "/"));
  var curMonth = curDate.getMonth();
  curDate.setMonth(curMonth + t + 1);
  curDate.setDate(0);
  var curDates = new Array(curDate.getDate()).fill(0).map((item, key) => {
    return key + 1;
  });
  return curDates;
};

const getWeek = (d) => {
  var curDate = new Date(d.replace(/-/g, "/"));
  return curDate.getDay();
};

const getChunkDates = (d) => {
  var curDates = getDates(d);
  var preDates = getDates(d, -1);
  var nexDates = getDates(d, 1);
  var curWeek = getWeek(d);
  curDates = curDates.map((i, k) => {
    return {
      num: i,
      fullDate: /(^\d{4})-(\d{1,2})-/.exec(d)[0] + i,
      show: 1,
    };
  });
  preDates = preDates.map((i) => {
    return {
      num: i,
      show: 0,
    };
  });
  nexDates = nexDates.map((i) => {
    return {
      num: i,
      show: 0,
    };
  });
  var preCurDates = preDates
    .slice(preDates.length - (curWeek === 0 ? 6 : curWeek - 1), preDates.length)
    .concat(curDates);
  return preCurDates.concat(nexDates.slice(0, 42 - preCurDates.length));
};

const get = (s) => {
  return isUniApp() ? uni.getStorageSync(s) : localStorage.getItem(s);
};

const set = (s, v) => {
  isUniApp() ? uni.setStorageSync(s, v) : localStorage.setItem(s, v);
};

const remove = (s) => {
  isUniApp() ? uni.removeStorageSync(s) : localStorage.removeItem(s);
};

var localCache = {
  tokenSet: (s) => {
    const { cache } = getGlobalVariable();
    set(cache?.tokenKey, s);
  },
  tokenGet: () => {
    const { cache } = getGlobalVariable();
    return get(cache?.tokenKey);
  },
  tokenRemove: () => {
    const { cache } = getGlobalVariable();
    remove(cache?.tokenKey);
  },
};

class Http {
  constructor() {
    this.queue = {};
  }

  getInstance() {
    return isUniApp() ? (options) => {
      return new Promise((resolve, reject) => {
        uni.request({
          ...options,
          success: (res) => {
            resolve(res.data);
          }
        });
      });
    } : (() => {
      const { axios } = getGlobalVariable('libs');
      return axios.create();
    })()
  }

  getAdaptOptions(options) {
    const { data, baseURL, url, params } = options;
    let adaptOptions = {};
    if (isUniApp()) {
      adaptOptions = {
        method: "POST",
        data,
        url: baseURL + url + stringifyQuery(params),
        header: {
          'Content-Type': 'application/json',
        }
      };
    } else {
      adaptOptions = {
        headers: {
          "Content-Type": "application/json",
        },
        method: "post",
        baseURL,
        url,
        params,
        data
      };
    }
    return adaptOptions;
  }

  destroy(url) {
    delete this.queue[url];
  }

  interceptors(instance, adaptOptions) {
    const { url } = adaptOptions;
    const { cache, hook } = getGlobalVariable();
    if (isUniApp()) {
      adaptOptions['header'] = {
        ...adaptOptions['header'],
        ...(() => {
          if (localCache.tokenGet()) {
            headers[cache?.jwtKey] = localCache.tokenGet();
          }
          return {}
        })()
      };
    } else {
      instance.interceptors.request.use(
        (config) => {
          if (hook?.httpBefore) {
            hook.httpBefore(config);
          }
          if (localCache.tokenGet()) {
            config.headers[cache?.jwtKey] = localCache.tokenGet();
          }
          this.queue[url] = true;
          return config;
        },
        (error) => {
          return Promise.reject(error);
        }
      );
      instance.interceptors.response.use(
        (res) => {
          if (hook?.httpAfter) {
            hook.httpAfter(res);
          }
          if (res.headers[cache?.jwtKey]) {
            localCache.tokenSet(res.headers[cache?.jwtKey]);
          }
          this.destroy(url);
          const { data } = res;
          return data;
        },
        (error) => {
          console.log(error);
          this.destroy(url);
          return Promise.reject(error);
        }
      );
    }
  }

  request(options) {
    const instance = this.getInstance();
    const adaptOptions = this.getAdaptOptions(options);
    this.interceptors(instance, adaptOptions);
    return instance(adaptOptions)
  }
}

const http = new Http();

const refUtil = () => {
	const _refMap = reactive(new Map());
	return {
		_refMap,
		storeRef: (el) => {
			if (el && el.$attrs['ref-key']) {
				_refMap.set(el.$attrs['ref-key'], el);
			}
		},
		getRef: (k) => {
			return _refMap.get(k);
		}
	}
};

const formToolkit = {
  validate: (obj, config = { rqeuire: [] }) => {
    const result = {
      ok: true,
      form: null,
    };
    const getVal = (f, obj, i) => {
      return _get(f, obj[i].keyPath ? `${obj[i].keyPath}.${i}` : `${i}`);
    };
    const setVal = (f, obj, i, v) => {
      _set(f, obj[i].keyPath ? `${obj[i].keyPath}.${i}` : `${i}`, v);
    };
    const _loop = (obj) => {
      if (Array.isArray(obj)) {
        return obj.map((i) => _loop(i));
      }
      const f = {};
      for (const i in obj) {
        if (i === "FILED") {
          continue;
        }
        if (obj[i].skip) {
          continue;
        }
        if (obj[i].require && !obj[i].value) {
          if (config.rqeuire?.length && !config.rqeuire?.includes(i)) {
            continue;
          }
          new Log4js().log(`validate error: key-------> ${i}`);
          obj[i]._validator = {
            ok: false,
            msg: obj[i].errorTip || "请校验输入信息",
          };
          result.ok = false;
        }
        if (obj[i].isJson) {
          setVal(
            f,
            obj,
            i,
            isProxy(obj[i].value) ? toRaw(obj[i].value) : obj[i].value
          );
          continue;
        }
        if (obj[i].toStr) {
          setVal(f, obj, i, JSON.stringify(obj[i].value));
          continue;
        }
        if (obj[i].skipFill && !obj[i].value) {
          continue;
        }
        if (Array.isArray(obj[i].value)) {
          if (!obj[i].value.length) {
            setVal(f, obj, i, []);
            continue;
          }
          obj[i].value.forEach((j) => {
            let isSkipRow = false;
            if (!f[i]) {
              setVal(f, obj, i, []);
            }
            for (const k in j) {
              if (
                j[k].skipRowValue !== undefined &&
                j[k].skipRowValue === j[k].value
              ) {
                isSkipRow = true;
              }
            }
            if (isSkipRow) {
              return;
            }
            getVal(f, obj, i)?.push(_loop(j));
          });
          continue;
        }
        if (
          typeof obj[i].value === "object" &&
          Object.keys(obj[i].value).length
        ) {
          setVal(f, obj, i, _loop(obj[i].value));
          continue;
        }
        setVal(f, obj, i, obj[i].value);
      }
      return f;
    };
    result.form = _loop(obj);
    return result;
  },
  fill: (form, obj) => {
    const { entities } = getGlobalVariable();
    for (const i in form) {
      if (i === "FILED") {
        continue;
      }
      if (form[i].skip) {
        form[i].value =
          obj[i] ||
          (() => {
            const type = form[i].type.split("|")[0];
            if (type === "array") {
              return [];
            }
            if (type === "string") {
              return "";
            }
            if (type === "boolean") {
              return false;
            }
            return {};
          })();
        continue;
      }
      if (form[i].skipFill) {
        continue;
      }
      if (form[i].toStr) {
        form[i].value = obj[i].toParse();
        continue;
      }
      const [typeValid, entity] = form[i].type.split(">");
      const [type] = typeValid.split("|");
      if (type === "array") {
        obj[i] &&
          obj[i].forEach((j) => {
            form[i].value.push(
              formToolkit.fill(
                entities[entity]
                  .formUtil()
                  .generator("update", "default"),
                j
              )
            );
          });
        continue;
      }
      if (type === "object") {
        form[i].value = formToolkit.fill(
          entities[entity].formUtil().generator("update", "default"),
          obj[i]
        );
        continue;
      }
      form[i].value = obj[i];
    }
    return form;
  },
};

const _baseTableField = {
  id: {
    type: "int",
    default: 0,
  },
  sn: {
    type: "int",
    default: 0,
  },
  status: {
    type: "int",
    default: 0,
  },
  isDelete: {
    type: "int",
    default: 0,
  },
  createTime: {
    type: "string",
    default: "",
  },
  updateTime: {
    type: "string",
    default: "",
  },
  deleteTime: {
    type: "string",
    default: "",
  },
};

class Entity {
  _tableField = [];
  request = null;
  _raw = null;
  constructor(entity, { _tableField }) {
    const { entities } = getGlobalVariable();
    this._tableField = Object.assign(_cloneDeep(_baseTableField), _tableField);
    for (const i in this._tableField) {
      if (entity && this._tableField[i].isEntity) {
        let v = _get(entity, i);
        v = (typeof v) === 'string' ? toParse(v || "{}") : v;
        const e = entities[this._tableField[i].type];
        if (e) {
          this[i] = Array.isArray(v) ? v.map((j) => new e(j)) : new e(v);
        }
        continue;
      }
      this[i] = entity
        ? this._fieldDo(entity, i)
        : _get(this._tableField, `${i}.default`);
      this._raw = entity;
    }
  }

  _fieldDo(entity, i) {
    if (!this._tableField[i].do) {
      return _get(entity, i);
    }
    const { parse } = this._tableField[i].do;
    if (parse) {
      return (typeof _get(entity, i)) === 'string' ? toParse(_get(entity, i) || "{}") : _get(entity, i);
    }
  }

  toEnum(key, dv = "--") {
    const tar = _find(this._remoteConfig?.enumValues, (x) => x.key === key);
    if (tar) {
      return _find(tar.list, (x) => x.value === this[key]);
    }
    return dv;
  }

  static async sendApi(
    method,
    form = { params: {}, body: {} },
    config = {
      errorTip: "",
      parse4Entity: false,
      showLoading: false,
      enableRemoteConfig: false,
    }
  ) {
    const { app, domain } = this._requestConfig;
    const { api: apiJSON } = getGlobalVariable('app');
    const { params, body } = form;
    let remoteConfig = null;
    let url = '';
    try {
      url = apiJSON[app].modules[domain][method].path;
    } catch (e) {
      // eslint-disable-next-line no-undef
      new Log4js().error(`
      request.js =====> api json configure is error
      full path ======> ${app}/${domain}/${method}
      `);
    }
    if (config.enableRemoteConfig) {
      const { url, transformer } = getGlobalVariable('remote')?.entity;
      const res = await http
        .request({
          baseURL: url,
          url: "",
          params: {
            domain: config.enableRemoteConfig,
          },
          data: {},
        })
        .finally(() => { });
      remoteConfig = transformer(res, 'config');
    }
    if (config.showLoading) ;
    return http.request({
      baseURL: apiJSON[app].hostname,
      url,
      params,
      data: body
    })
      .then((res) => {
        const { status, message } = res;
        if (!status) {
          if (config.errorTip) {
            if (typeof config.errorTip === "function") {
              config.errorTip(res);
            }
          }
        }
        if (config.parse4Entity) {
          if (Array.isArray(res.data)) {
            res.data = res.data.map((i) => {
              const _i = new this(i);
              _i._remoteConfig = remoteConfig;
              return _i;
            });
          } else {
            const _i = new this(res.data);
            _i._remoteConfig = remoteConfig;
            res.data = _i;
          }
        }
        return res;
      });
  }

  static formUtil() {
    return {
      generator: (module, type) => {
        const f = _get(this._form, `${module}.${type}`);
        for (const i in f) {
          f[i] = Object.assign(f[i], { _validator: { ok: true, msg: "" } });
        }
        return {
          ..._cloneDeep(f),
          FILED(s) {
            const _self = this;
            let k = "";
            const sr = s.split(".");
            for (let i in sr) {
              const r = sr[i].match(/(\S+)\[(\d+)\]$/);
              if (r && r.length === 3) {
                k = k + `${r[1]}.value[${r[2]}]`;
                continue;
              }
              k =
                k + (i == sr.length - 1 ? `${sr[i]}.value` : `${sr[i]}.value.`);
            }
            return {
              get val() {
                return _get(_self, k);
              },
              set val(v) {
                _set(_self, k, v);
              },
              get valid() {
                return _get(_self, k.slice(0, k.length - 6));
              },
            };
          },
        };
      },
      validate: (json) => {
        return formToolkit.validate(json);
      },
      fill: (f, json, config) => {
        for (const i in f) {
          f[i] = Object.assign(f[i], { _validator: { ok: true, msg: "" } });
        }
        return formToolkit.fill(f, json, config);
      },
    };
  }
}

export { Entity, Log4js, _cloneDeep, _find, _get, _set, emitter, formToolkit, genArrayByLength, getChunkDates, getDates, getGlobalVariable, getObjVal, getWeek, http, isObjHasBlank, isUniApp, refUtil, setGlobalVariable, stringifyQuery, toParse, toStringify, uuid };
